import {action, computed, makeObservable, observable, reaction, runInAction} from "mobx";
import {
    createOrderValidationSchema,
    OrderModel
} from "../stores/OrderStore";
import {
    Field,
    Fields,
    FrameCandidate,
    FrameFindFilters,
    FrameFindOption, FrameOnlyFields,
    FrameTrace, IncludedTraceKeys,
    JOB_TYPES,
    LensUseIndicator, prismOptions,
    ProcessedOption,
    RawFrameFindOption, TraceKeyMap,
    UnprocessedOption
} from "../stores/_FieldConstants";
import {mixed, number, object, Schema, string} from "yup";
import {t} from "@lingui/macro";
import {errToString, FetchError, schemas, Transport} from "@rx-suite/data";
import {moduleIsInstalled} from "@rx-suite/stores";
import {computedFn, createViewModel} from "mobx-utils";
import * as moment from "moment/moment";
import {CurrentUserConfig} from "../stores/ConfigStore";
import {customerSelectionStore} from "./useCustomerSelectionStore";

// https://stackoverflow.com/questions/3966484/why-does-modulus-operator-return-fractional-number-in-javascript/31711034#31711034
function floatSafeRemainder(val: number, step: number) {
    const valDecCount = (val.toString().split('.')[1] || '').length;
    const stepDecCount = (step.toString().split('.')[1] || '').length;
    const decCount = valDecCount > stepDecCount ? valDecCount : stepDecCount;
    const valInt = parseInt(val.toFixed(decCount).replace('.', ''));
    const stepInt = parseInt(step.toFixed(decCount).replace('.', ''));
    return (valInt % stepInt) / Math.pow(10, decCount);
}

const generateSchema = (genName: string, f: Field) => {
    const label = f.desc || genName;
    let numberSchema = number().transform((value, original) => {
        if (original === "") {
            return undefined;
        } else {
            return value;
        }
    }).min(Number(f.minValue), t`${label} must be greater than or equal to ${f.minValue}`).max(Number(f.maxValue), t`${label} must be less than or equal to ${f.maxValue}`).test({
        name: "step",
        test: n => n === undefined ? false : floatSafeRemainder(n, Number(f.increment)) === 0,
        message: t`${label} must be increments of ${f.increment}`,
        skipAbsent: true
    }).typeError(t`${label} must be a number`);

    const options = f.options?.map(o => o.code) || [];
    const codes = [...options];

    if (f.mandatory === 1) {
        numberSchema = numberSchema.required(`${genName} is a required field`); // it's a number and required, we can just mark the schema as required
    } else {
        // @ts-ignore the type doesn't handle null, but it's valid
        codes.push("", null); // options are implicitly required because it _must_ equal one of the options. If it's not required, make an empty string a valid option.
    }
    if (f.name == "PRVA1" || f.name == "PRVA2") {
        codes.push(...prismOptions.filter(o => !codes.includes(o)))
    }
    const optionsSchema = string().oneOf(codes, t`${genName} must be one of the following values: ${codes.join(", ")}`).nullable();

    const validateOptions = (n: string, errors: string[]) => {
        try {
            optionsSchema.validateSync(n);
        } catch (e) {
            errors.push(e.message);
        }
    }

    const validateNumber = (n: string, errors: string[]) => {
        try {
            numberSchema.validateSync(n);
        } catch (e) {
            errors.push(e.message);
        }
    }

    const validateCyl = (n: string, other: string, errors: string[]) => {
        if(!n) return true;
        if(!other) return true;
        if(n === "0") return true;
        if(other === "0") return true;

        const signA = Math.sign(parseFloat(n));
        const signB = Math.sign(parseFloat(other));
        if(signA !== signB) {
            errors.push(t`Cylinders must both be positive, or negative`)
        }
    }

    return mixed().nullable().test({
        name: '💩',
        test: (n, ctx) => {

            const errors: string[] = [];

            if(genName === "rtCYL") {
                validateCyl(n, ctx.parent["ltCYL"], errors);
            }
            if(genName === "ltCYL") {
                validateCyl(n, ctx.parent["rtCYL"], errors);
            }

            // if the field is a number
            if (f.increment) {
                validateNumber(n, errors);

                if (errors.length) { // if it's not a number
                    if (options.length) { // if it could be an option
                        validateOptions(n, errors);
                        if (errors.length > 1) { // if it's also not an option
                            const errorMessage = errors.join(" or ");

                            return ctx.createError({
                                message: errorMessage
                            })

                        } else { // it's a valid option
                            return true;
                        }
                    }

                    // it's not a number and can't be an option
                    return ctx.createError({
                        message: errors[0]
                    });
                }
            } else if (options.length) { // it must be an option
                validateOptions(n, errors);
                if (errors.length) { // it's not an option
                    return ctx.createError({
                        message: errors[0]
                    })
                }
            }

            // nothing else has errored
            return true
        }
    })
}

const maybeThrowResponse = async (response: Response) => {
    if (response.ok) {
        return response;
    } else {
        throw new FetchError(await response.json());
    }
}

export const getLeftName = (rightName: string) => {
    if (rightName.startsWith('raw_data.')) {
        return rightName.replace(".rt", ".lt");
    } else if (rightName.startsWith('rt')) {
        return `lt${rightName.slice(2)}`
    } else {
        return `lt${rightName}`;
    }
};

class OrderEntryStore {
    @observable order: OrderModel | null = null;

    @observable fetchingFields = false;
    @observable fetchingFieldsError = '';
    @observable perEye: Field[] = [];
    @observable perJob: Field[] = [];
    @observable leftEyeMode: LensUseIndicator = 'USE';
    @observable rightEyeMode: LensUseIndicator = 'USE';

    viewModel: OrderEntryStore;

    constructor() {
        makeObservable(this);
        this.viewModel = createViewModel(this);
    }

    @action.bound
    async setTarget(order: OrderModel, readOnly = false) {
        this.order = order;
        this.order.readOnly = readOnly;

        const currentUserConfig = CurrentUserConfig.get();



        if(order.isClone) {
            order.nestedAssignment("order_number", "");
            order.nestedAssignment("date_ordered", moment());
            order.nestedAssignment("raw_data.date_wanted", undefined);
            order.nestedAssignment("unvalid_message", "");
        } else if(order.isNew) {
            orderEntryStore.frameTraceFile = null
            orderEntryStore.resetTrace();

            this.order?.nestedAssignment("customer_id", parseInt(customerSelectionStore.currentCustomerId))
            if (currentUserConfig) {
                this.setFrameStatus(currentUserConfig.default_frame_status);
                this.setJobType(currentUserConfig.default_job_type);
            }
        }
    }

    @action.bound
    async clear() {
        this.order = null;
        await this.clearFields();
    }

    @action.bound
    async clearFields() {
        this.perEye = [];
        this.perJob = [];
        if(this.order) {
            this.order.raw_data = {};
        }
        await this.resetSchema();
    }

    @action.bound
    async resetSchema() {
        console.log("Resetting schema", this.order);
        schemas["OrderModel"] = createOrderValidationSchema();
        await this.order?.validate()
    }

    @computed get hasFields() {
        return !!this.perEye.length || !!this.perJob.length;
    }

    @action.bound
    setRightEyeUsage(val: LensUseIndicator) {
        this.rightEyeMode = val;
    }

    @action.bound
    setLeftEyeUsage(val: LensUseIndicator) {
        this.leftEyeMode = val;
    }

    @computed
    get rightEnabled() {
        return (this.rightEyeMode === 'USE') || (this.rightEyeMode === 'BALANCE');
    }

    @computed
    get rightVisible() {
        return (this.rightEyeMode !== 'HIDE');
    }

    @computed
    get leftEnabled() {
        return (this.leftEyeMode === 'USE') || (this.leftEyeMode === 'BALANCE');
    }

    @computed
    get leftVisible() {
        return (this.leftEyeMode !== 'HIDE');
    }

    @action.bound
    setFrameOnlyFields() {
        this.perJob = FrameOnlyFields;

        if(this.order?.readOnly) return;

        const obj: { [index: string]: Schema } = {};

        FrameOnlyFields.forEach((f: Field) => {
            const genName = f.name;

            if(!this.order?.readOnly) {
                this.order!.raw_data[genName] = f.initialValue;
            }

            obj[genName] = generateSchema(genName, f);
        })

        // @ts-ignore
        schemas["OrderModel"] = createOrderValidationSchema().concat(object().shape({raw_data: object().shape(obj)}));

        this.order?.validate();
    }

    @action
    async fetchFields() {
        if(this.fetchingFields) return;

        const order = this.order;
        if(!order) return;

        const styleId = order.viewModel.dig("style_id");
        const frameStatus = order.viewModel.dig("_fstatus");

        try {
            this.fetchingFields = true;
            this.fetchingFieldsError = '';

            const response = await Transport.fetch(`/api/fields?style=${styleId}&frame_status=${frameStatus}`);

            if(order !== this.order) {
                console.warn("Order changed while fetching fields, trying again...");

                runInAction(() => {
                    this.fetchingFields = false;
                    this.tryFetchFields();
                })


                return;
            }

            await maybeThrowResponse(response);
            const json = await response.json() as Fields;

            runInAction(() => {

                const perEye = json.PerEye;
                const perJob = json.PerJob.filter((f: Field) => {
                    if (f.name == "TraceID"){
                        f.name = "traceID";
                    }
                    return f.name !== '_FSTATUS'; // frame status is handled specially
                });

                this.perEye = perEye;
                this.perJob = perJob;

                if(order.readOnly) return;

                const obj: { [index: string]: Schema } = {};

                perEye.forEach((f: Field) => {
                    const eyes = [] as string[];
                    if (this.rightEnabled) {
                        eyes.push('rt');
                    }
                    if (this.leftEnabled) {
                        eyes.push('lt');
                    }
                    eyes.forEach(eye => {
                        const genName = eye + f.name;

                        if (!order.readOnly) {
                            order.raw_data[genName] = f.initialValue;
                        }

                        obj[genName] = generateSchema(genName, f);
                    })
                });

                perJob.forEach((f: Field) => {
                    const genName = f.name;

                    if (!order.readOnly) {
                        order.raw_data[genName] = f.initialValue;
                    }

                    obj[genName] = generateSchema(genName, f);
                })

                console.log("Setting schema")

                if(order.readyOnly) {
                    this.resetSchema();
                } else {
                    // @ts-ignore
                    schemas["OrderModel"] = createOrderValidationSchema().concat(object().shape({raw_data: object().shape(obj)}));
                }


                order.validate();
            })
        } catch (e) {
            runInAction(() => this.fetchingFieldsError = errToString(e));
        } finally {
            runInAction(() => this.fetchingFields = false);
        }
    }

    @action
    async tryFetchFields() {
        const styleId = this.order?.viewModel.dig("style_id");

        if(styleId) {
            await this.fetchFields();
        } else if(this.order?.readOnly) {
            await this.clearFields();
        }
    }

    @observable showingFrameDialog = false;
    @observable fetchingFrameCandidates = false;
    @observable fetchingFrameCandidatesError = '';
    @observable possibleFrames: FrameCandidate[] = [];

    @observable hasLoadedFrameFindOptions = false;
    @observable fetchingFrameFindOptions = false;
    @observable fetchingFrameFindOptionsError = '';
    @observable frameFindOptions: FrameFindOption[] = [];
    @observable frameFindFilters: FrameFindFilters = {};

    @observable frameTrace: FrameTrace | null = null;
    @observable fetchingFrameTrace = false;
    @observable fetchingFrameTraceError = '';
    @observable frameTraceFile: any  | null  = null;
    @observable loadingFrameTraceFile = false;

    @observable upcLookupFieldValue = '';

    maybeGetFrameFinderOptions() {
        if (!this.hasLoadedFrameFindOptions && !this.fetchingFrameFindOptions) {
            this.getFrameFinderOptions();
        }
    }

    @action.bound resetTrace() {
        if(this.order?.readOnly) return;

        IncludedTraceKeys.forEach((value => {
            this.updateRawDataValue(value, '');
        }))
        this.fetchingFrameTraceError = '';
    }

    @action.bound resetLensInformation() {
        this.order?.nestedAssignment("quick_code_id",'')
        this.order?.nestedAssignment("style_id",'')
        this.order?.nestedAssignment("style_material_id",'')
        this.order?.nestedAssignment("style_type_id",'')
    }

    @action.bound
    async getFrameFinderOptions() {
        if (moduleIsInstalled('rx_inventory') && !this.fetchingFrameFindOptions) {
            try {
                runInAction(() => {
                    this.fetchingFrameFindOptions = true;
                    this.fetchingFrameFindOptionsError = '';
                });
                const response = await Transport.fetch('/api/rx_inventory_adapter/frame_search_options');
                await maybeThrowResponse(response);
                const json = await response.json() as RawFrameFindOption[];
                const processed = json.map(o => {
                    return {
                        label: o.label,
                        id: o.id,
                        corresponds_to: o.corresponds_to,
                        options: o.options.map((maybeString: UnprocessedOption) => {
                            if (typeof maybeString === 'string') {
                                return {
                                    label: maybeString,
                                    value: maybeString
                                }
                            } else {
                                return maybeString as ProcessedOption;
                            }

                        })
                    } as FrameFindOption;
                });
                runInAction(() => {
                    this.frameFindOptions = processed;
                    this.hasLoadedFrameFindOptions = true;
                });
            } catch (e) {
                runInAction(() => this.fetchingFrameFindOptionsError = errToString(e));
            } finally {
                runInAction(() => {
                    this.fetchingFrameFindOptions = false;
                });
            }
        }
    }

    frameFindFilterValue = computedFn((id: string) => {
        return this.frameFindFilters[id] || '';
    });

    @action.bound
    handleFrameFindFilterChange(id: string | undefined, value: string) {
        if (id) {
            this.frameFindFilters = {...this.frameFindFilters, [id]: value};
        }
    }

    @action.bound
    handleFrameUPCChange(e: any) {
        this.upcLookupFieldValue = e.target.value
    }

    @action.bound
    showFrameDialog() {
        this.showingFrameDialog = true;
    }

    @action.bound
    hideFrameDialog() {
        this.showingFrameDialog = false;
    }

    @action.bound
    toggleShowingFrameDialog() {
        this.showingFrameDialog = !this.showingFrameDialog;
    }

    @action.bound
    async maybeCommitUPCLookupField() {
        if (this.upcLookupFieldValue) {
            this.commitUPCLookupField();
        }
    }

    @action.bound
    async commitUPCLookupField() {
        await this.findFramesByUPC(this.upcLookupFieldValue);
        runInAction(() => {
            this.upcLookupFieldValue = '';
        });
    }

    @action.bound
    handleDrop(files: any) {
        this.loadingFrameTraceFile = true
        this.frameTraceFile = files[0] || null;
        if(this.frameTraceFile) {
            this.loadingFrameTraceFile = true;
            this.resetTrace();

            const reader = new FileReader();
            reader.addEventListener('load', event => {
                if (event.target) {
                    this.loadingFrameTraceFile = false;
                    if (reader.result){
                        try {
                            this.parseTraceFile(reader.result as string);

                        } catch (err) {
                            this.fetchingFrameTraceError = "Error in uploading trace file"
                        }
                    } else {
                        this.fetchingFrameTraceError = "The uploaded trace file is empty"
                    }

                }
            });
            reader.readAsText(this.frameTraceFile);
        }
        this.loadingFrameTraceFile = false;
    }

    @action.bound
    async findFramesByUPC(upc: string) {
        if (upc && moduleIsInstalled('rx_inventory') && !this.fetchingFrameCandidates) {
            runInAction(() => {
                this.fetchingFrameCandidates = true;
                this.fetchingFrameCandidatesError = '';
            });
            try {
                const response = await Transport.fetch(`/api/rx_inventory_adapter/frames?upc=${upc}`);
                await maybeThrowResponse(response);
                const json = await response.json() as FrameCandidate[];
                runInAction(() => {
                    this.possibleFrames = json;
                    this.resolveFrameSelection();
                });
            } catch (e) {
                runInAction(() => {
                    this.fetchingFrameCandidatesError = errToString(e);
                });
            } finally {
                runInAction(() => {
                    this.fetchingFrameCandidates = false;
                });
            }
        } else if (!this.fetchingFrameCandidates) {
            runInAction(() => {
                this.possibleFrames = [];
                this.fetchingFrameCandidatesError = '';
            });
        }
    }

    @action.bound
    async findFrames() {
        if (this.hasFrameFilter && moduleIsInstalled('rx_inventory') && !this.fetchingFrameCandidates) {
            runInAction(() => {
                this.fetchingFrameCandidates = true;
                this.fetchingFrameCandidatesError = '';
            });
            try {
                const url = new URL('/api/rx_inventory_adapter/find_frames', window.location.origin);
                Object.entries(this.frameFindFilters).forEach(([id, val]) => {
                    if (val) {
                        url.searchParams.append(id, val);
                    }
                });
                const response = await Transport.fetch(url.toString());
                await maybeThrowResponse(response);
                const json = await response.json() as FrameCandidate[];
                runInAction(() => {
                    this.possibleFrames = json;
                });
            } catch (e) {
                runInAction(() => {
                    this.fetchingFrameCandidatesError = errToString(e);
                });
            } finally {
                runInAction(() => {
                    this.fetchingFrameCandidates = false;
                });
            }
        } else if (!this.fetchingFrameCandidates) {
            runInAction(() => {
                this.possibleFrames = [];
                this.fetchingFrameCandidatesError = '';
            });
        }
    }

    @action.bound
    resolveFrameSelection() {
        if (this.possibleFrames.length === 1) {
            this.applyFrame(this.possibleFrames[0]);
        } else if (this.possibleFrames.length > 1) {
            this.showFrameDialog();
        } else {
            this.fetchingFrameCandidatesError = t`No frames found`;
        }
    }

    @action.bound
    applyFrame(frame: FrameCandidate) {
        if (this.order && frame) {
            const trace = frame.trace || ''
            this.order.viewModel.raw_data = {
                ...this.order.viewModel.raw_data,
                _FMUPC: frame.manufacturer_upc || '',
                FRAM: frame.name || '',
                FUPC: frame.upc || '',
                FCOL: frame.color || '',
                traceID: trace,
                HBOX: frame.a?.toString() || '',
                VBOX: frame.b?.toString() || '',
                FED: frame.ed?.toString() || '',
                DBL: frame.bridge?.toString() || '',
                _ETYPE: frame.edge_type,
            };
            this.order.viewModel.frame_sku = frame.id;
            this.order.trace_id = trace;
            this.order.fram = frame.name || '';
            this.order.fupc = frame.upc || '';
            this.order._fmupc = frame.manufacturer_upc || '';

            this.order?.validate();
        }
        this.hideFrameDialog()
    }

    @action.bound
    async getTrace(id: string) {
        if (id && !this.fetchingFrameTrace) {
            runInAction(() => {
                this.fetchingFrameTrace = true;
                this.fetchingFrameTraceError = '';
            });
            try {
                const response = await Transport.fetch(`/api/rx_inventory_adapter/frame_traces/${encodeURIComponent(id)}`);
                await maybeThrowResponse(response);
                const json = await response.json() as FrameTrace;

                runInAction(() => {
                    if(json === null) {
                        this.fetchingFrameTraceError = t`Could not find trace ${id} in RxUniverse`
                    } else {
                        this.frameTrace = json;

                        if (this.order) {
                            this.order.nestedAssignment('raw_data.HBOX',this.frameTrace.a.toString());
                            this.order.nestedAssignment('raw_data.VBOX',this.frameTrace.b.toString());
                            this.order.nestedAssignment('raw_data.FED',this.frameTrace.ed.toString());
                            this.order.nestedAssignment('raw_data.CIRC',this.frameTrace.circ.toString());
                            this.order.nestedAssignment('raw_data.DBL',this.frameTrace.dbl.toString());
                            this.order.nestedAssignment('trace_id',this.frameTrace.name.toString());
                        }
                    }
                })
            } catch (e) {
                runInAction(() => {
                    this.fetchingFrameTraceError = errToString(e);
                });
            } finally {
                runInAction(() => {
                    this.fetchingFrameTrace = false;
                    this.order?.validate();
                });
            }
        }
    }

    @action.bound
    setFrameStatus(frameStatus: string) {
        this.order?.nestedAssignment("_fstatus", frameStatus);
        if(orderEntryStore.mustPickFrame){
            this.frameTraceFile = null
            this.resetTrace();
        }
    }

    @action.bound
    setTraceId(traceId: string) {
        this.order?.nestedAssignment("raw_data.traceID", traceId);
        if(traceId) {
            this.getTrace(traceId);
            this.frameTraceFile = null;
        } else {
            this.resetTrace();
        }
    }
    @action.bound
    setJobType(job_type: string) {
        orderEntryStore.order?.nestedAssignment("job_type", job_type);

        const jobType = JOB_TYPES.find(t => t.value === job_type);

        console.log(jobType.value);
        if (jobType && orderEntryStore.order) {
            if (jobType.value == "A"){
                this.resetLensInformation();
            } else if (jobType.value == "F") {
                this.resetLensInformation();
                this.setFrameOnlyFields();
            } else {
                this.setLeftEyeUsage(jobType.left);
                this.setRightEyeUsage(jobType.right);

                if (!this.order?.isFrameOnlyJob) {
                    this.clearFields();
                }
            }
        } else {
            console.warn(`Unrecognized job type: ${job_type}`);
        }
    }

    @action.bound
    async parseTraceFile(trace: string | null | undefined) {
        if (trace){
            try {
                const uploadedTrace = {} as FrameTrace;
                const traceArray = (trace.split(/[\s]+/) || []).map(line => {
                    const path = line.split("=") || [];
                    return {key: path[0].toUpperCase(), value: path[1]}
                }).filter( a => IncludedTraceKeys.includes(a.key) || a.key === "TRCFMT");

                const numberOfRadiiPerEye = (traceArray.find( a => a.key == "TRCFMT") || {}).value?.split(";")[1]  || -1
                const allRadiiPoints = traceArray.reduce(
                    (previousValue: number[], currentValue) => {
                        if (currentValue.key == "R") {
                            const points = currentValue.value.split(";") || []
                            const parsedPoints = points.map(a => parseFloat(a)/100);
                            return [...previousValue,...parsedPoints]
                        }
                        return previousValue;
                    },
                    [],
                );

                const singleEyeRadiiPoints = numberOfRadiiPerEye != -1 ? allRadiiPoints.slice(0, parseInt(numberOfRadiiPerEye) - 1 ) : [];

                uploadedTrace["radii"] = singleEyeRadiiPoints;
                traceArray.filter(a => a.key != "R").forEach((traceData) => {
                    const correctedValue = (traceData.value?.split(";") || [])[0];
                    if(correctedValue && this.order){
                        this.updateRawDataValue(traceData.key, correctedValue);
                        uploadedTrace[TraceKeyMap[traceData.key]] = correctedValue
                    }
                });

                runInAction(() => this.frameTrace = uploadedTrace);

            } catch (e) {
                runInAction(() => {
                    this.fetchingFrameTraceError = errToString(e);
                });
            } finally {
                runInAction(() => {
                    if (this.order) {
                        this.order?.validate();
                    }
                });
            }
        }
    }

    @computed
    get hasFrameFilter() {
        return Object.values(this.frameFindFilters).some(f => !!f);
    }

    @computed
    get canPickFrame() {
        return this.order?.viewModel?._fstatus !== "D"
    }

    @computed
    get mustPickFrame() {
        return this.order?.viewModel?._fstatus === "S"
    }

    @computed
    get lockFrameFields() {
        return this.order?.readOnly || this.mustPickFrame || this.fetchingFrameCandidates;
    }

    updateRawDataValue(key : string, value : any){
        (this.order as OrderModel).nestedAssignment(`raw_data.${key}`, value.toString());
    }
}

const orderEntryStore = new OrderEntryStore();
makeObservable(orderEntryStore);

reaction(() => [orderEntryStore.order?.shipping_methods_id, orderEntryStore.order?.readOnly], () => {
    if(orderEntryStore.order?.shipping_methods_id == null){
        orderEntryStore.order?.nestedAssignment("shipping_methods_id", "");
    }
})

// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
if (!window.rxstores) {
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    window.rxstores = {};
}
// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
if (!window.rxstores.order_entry_store) {
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    window.rxstores.order_entry_store = orderEntryStore;
}

export const useOrderEntryStore = () => {
    return orderEntryStore;
}
