import * as React from 'react';
import {BreadModel, makeColumn, CheckboxField} from "@rx-suite/ui";
import {OrderForm} from "./OrderForm";
import {t, Trans} from "@lingui/macro";
import {OrderStore} from "../../stores/OrderStore";
import {DataStore, Model} from '@rx-suite/data';
import {Tab, Tabs} from "@material-ui/core";
import {useCustomerSelectionStore} from "../../hooks/useCustomerSelectionStore";
import {observer} from "mobx-react-lite";
import {Alert} from "@material-ui/lab";
import Typography from "@material-ui/core/Typography";
import {mappedFrameStatus, mappedJobType, mappedShippingMethod} from "../../stores/_FieldConstants";
import {Filter, RowInfo} from "react-table";
import CSS from "csstype";
import {red, amber} from "@material-ui/core/colors";
import {useMemo} from "react";
import {RxUniverseStore} from "../../stores/RxUniverseStore";

const getTrProps = (_: any, rowInfo: RowInfo | undefined) => {
    const props: {style?: CSS.Properties} = {};
    if (rowInfo && rowInfo.original && rowInfo.original.isInvalidJob) {
        props.style = {color: red[500], fontWeight: "bold"};
    }
    if (rowInfo && rowInfo.original && rowInfo.original.hasComment && !rowInfo.original.isInvalidJob) {
        props.style = {color: amber[500], fontWeight: "bold"}
    }
    return props;
};

export const OrdersBreadModel = observer(() => {
    const customerSelectionStore = useCustomerSelectionStore();
    const store = OrderStore as unknown as DataStore<Model>;

    const [tab, setTab] = React.useState<"summary" | "details">("summary");
    const [allAccounts, setAllAccounts] = React.useState(false);
    const [invalidJobs, setInvalidJobs] = React.useState(false);
    const [defaultFiltered, setDefaultFiltered] = React.useState<Array<Filter>>([{id: 'customer_id', value: customerSelectionStore.currentCustomerId}]);

    const releaseLabel = RxUniverseStore.selectedModel.isVA ? t`Last 4` : t`Release Number`;

    const detailsColumns = useMemo(() => {
        return [
            makeColumn('order_number', t`Order Number`, "string"),
            makeColumn('customer_num', t`Customer Number`, "string"),
            makeColumn('customer_name', t`Customer Name`, "string"),
            makeColumn('customer_city', t`Customer City`, "string"),
            makeColumn('date_received', t`Date Received`, "date"),
            makeColumn('date_ordered', t`Date Ordered`, "date"),
            makeColumn('cust_job_id', "Customer Job ID", "string"),
            makeColumn('po_number', t`PO Number`, "string"),
            makeColumn('release_number', releaseLabel, "string"),
            makeColumn('hold_at_lab', t`Hold At Lab?`, "boolean"),
            makeColumn('tray', t`Tray`, "string"),
            makeColumn('ship_to_name', t`Ship To Name`, "string"),
            makeColumn('ship_to_address1', t`Ship To Address1`, "string"),
            makeColumn('ship_to_address2', t`Ship To Address2`, "string"),
            makeColumn('ship_to_address3', t`Ship To Address3`, "string"),
            makeColumn('ship_to_address4', t`Ship To Address4`, "string"),
            makeColumn('shipping_methods_id', t`Shipping Method`, "string", mappedShippingMethod),
            makeColumn('job_type', t`Job Type`, "string", mappedJobType),
            makeColumn('style_type_name', t`Lens Type`, "string"),
            makeColumn('style_name', t`Lens Style`, "string"),
            makeColumn('style_material_name', t`Lens Material`, "string"),
            makeColumn('lens_color', t`Lens Color`, "string"),
            makeColumn('lens_treat', t`Lens Treat`, "string"),
            makeColumn('_fstatus', t`Frame Status`, "string", mappedFrameStatus),
            makeColumn('fupc', t`Frame UPC`, "string"),
            makeColumn('fram', t`Frame`, "string"),
            makeColumn('_fmupc', t`Frame MUPC`, "string"),
            makeColumn('_fsku', t`Frame SKU`, "string"),
            makeColumn("trace_id", t`Trace ID`, "string"),
            makeColumn('submitted_by', t`Submitted By`, "string")
        ];
    }, [releaseLabel, mappedShippingMethod, mappedJobType, mappedFrameStatus])

    const summaryColumns = React.useMemo(() => {
        return [
            makeColumn('order_number', t`Order Number`, "string"),
            makeColumn('customer_name', t`Patient Name`, "string"),
            makeColumn('release_number', releaseLabel, "string"),
            makeColumn('tray', "Tray", "string"),
            makeColumn('_fstatus', "Frame Status", "string", mappedFrameStatus),
            makeColumn('raw_data.station', "Job Track Station", "string"),
            makeColumn("raw_data.stationCustDesc", "Job Track Description", "string"),
            makeColumn('date_received', t`Date Received`, "date"),
            makeColumn('submitted_by', t`Submitted By`, "string")
        ]
    }, [releaseLabel, mappedFrameStatus])
    const tabs = {
        summary: summaryColumns,
        details: detailsColumns
    }

    const columns = tabs[tab];

    const handleAllAccountsChange = React.useCallback(e => {
        setAllAccounts(e.target.checked);
    }, [setAllAccounts]);

    const handleTabChange = React.useCallback((_, t) => {
        setTab(t);
    }, [setTab]);

    const handleInvalidJobChecked = React.useCallback(e => {
        setInvalidJobs(e.target.checked);
    }, [setInvalidJobs])

    React.useEffect(() => {
        const filter = [];

        if(allAccounts) {
            filter.push({id: 'customer_id', value: customerSelectionStore.userCustomers.map(c => c.getId()).join("|")})
        } else {
            filter.push({id: 'customer_id', value: customerSelectionStore.currentCustomerId})
        }

        if(invalidJobs) {
            filter.push({id: 'id', value: store.models.filter(j => j.isInvalidJob).map(o => o.getId()).join("|")})
        }

        setDefaultFiltered(filter);
    }, [allAccounts, invalidJobs])

    return (<>
            {customerSelectionStore.noCustomers && <Alert severity="error">
                <Typography variant={'h6'}>
                    <Trans>This account is not associated with a customer. You will be unable to create orders. Please
                        contact an administrator.</Trans>
                </Typography>
            </Alert>}
            <BreadModel store={store} Form={OrderForm} columns={columns}
                        tableProps={{
                            getTrProps: getTrProps,
                            defaultFiltered: defaultFiltered,
                            preTableActions: <>
                                <Tabs value={tab} onChange={handleTabChange}>
                                    <Tab label={t`Summary`} value={'summary'}/>
                                    <Tab label={t`Details`} value={'details'}/>
                                </Tabs>
                                <CheckboxField
                                    label={t`Include All Accounts?`}
                                    onChange={handleAllAccountsChange}
                                    checked={allAccounts}
                                />
                                <CheckboxField
                                    label={t`Show Only Invalid Jobs?`}
                                    onChange={handleInvalidJobChecked}
                                    checked={invalidJobs}
                                />
                            </>
                        }}/>
        </>
    )
});
