import {SelectField, useModelContext, useModelFieldProps, useReadonlyContext} from '@rx-suite/ui';
import {observer} from 'mobx-react-lite';
import {t} from '@lingui/macro';
import * as React from 'react';
import {LabMaintenanceStore} from "../../../stores/LabMaintenanceStore";
import {JobTypeCode, FrameStatus} from "../../../stores/_FieldConstants";
import {useOrderEntryStore} from "../../../hooks/useOrderEntryStore";

const UNCUT: FrameStatus = {value: "D", label: t`Uncut no Frame`};
const UNCUT_WITH_FRAME: FrameStatus = {value: "U", label: t`Uncut with Frame`};
const TRACE_ONLY: FrameStatus = {value: "L", label: t`Trace Only`};
const FRAME_TO_FOLLOW: FrameStatus = {value: "F", label: t`Frame to Follow`};
const LAB_SUPPLIED_FRAME: FrameStatus = {value: "S", label: t`Lab Supplied Frame`};
const FRAME_ENCLOSED: FrameStatus = {value: "E", label: t`Frame to Enclosed`};
const NONE: FrameStatus = {value: "N", label: t`None`};

const OPTIONS_MAP: { [key: string]: FrameStatus[] } = {
    P: [UNCUT, UNCUT_WITH_FRAME, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME,  NONE],
    R: [UNCUT, UNCUT_WITH_FRAME, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME,  NONE],
    L: [UNCUT, UNCUT_WITH_FRAME, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME, NONE],
    E: [NONE, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME],
    F: [LAB_SUPPLIED_FRAME],
    A: [],
    I: [],
    RB: [UNCUT, UNCUT_WITH_FRAME, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME, NONE],
    LB: [UNCUT, UNCUT_WITH_FRAME, TRACE_ONLY, FRAME_TO_FOLLOW, FRAME_ENCLOSED, LAB_SUPPLIED_FRAME, NONE],
    default: []
}

export const FrameStatusSelect = observer(() => {
    const model = useModelContext();
    const readOnly = useReadonlyContext();
    const orderEntryStore = useOrderEntryStore();

    //Frame Enclosed status is a readOnly status
    const options = React.useMemo(() => {
        let options: FrameStatus[] = OPTIONS_MAP[model.viewModel.job_type as JobTypeCode] || [];
        if (!readOnly){
           options = options.filter(a => a.value !== "E");
        }
        if(!LabMaintenanceStore.selectedModel?.display_uncut_orders) {
            options = options.filter(a => a.value !== "D")
        }
      return options;
    }, [readOnly, model.viewModel.job_type])

    // If the job type changes, make sure we're not selecting an invalid frame status for the job type
    React.useEffect(() => {
        const options = OPTIONS_MAP[model.viewModel.job_type as JobTypeCode] || [];
        const currentFrameStatus = model.viewModel._fstatus;

        const validOption = options.some((option: { value: string, label: string }) => {
            return option.value == currentFrameStatus;
        });

        if (!validOption) {
            if (options.length) {
                orderEntryStore.setFrameStatus(options[0].value);
            } else {
                orderEntryStore.setFrameStatus("");
            }

        }
    }, [model.viewModel, model.viewModel.job_type, orderEntryStore])

    const handleChange = React.useCallback((name: string, value: string) => {
        orderEntryStore.setFrameStatus(value);
    }, [orderEntryStore]);

    const selectProps = useModelFieldProps({ name: "_fstatus" })

    // Addons only don't have a frame
    if (model.isAddonJob) {
        return null;
    }

    return (
        <SelectField
            value={model.viewModel.dig("_fstatus") || null}
            onChange={handleChange}
            options={options}
            {...selectProps}
        />
    )
});
