import {attribute, DataStore, Model, model, schemas, Types} from '@rx-suite/data';
import {t} from "@lingui/macro";
import {string, array, object} from 'yup';
import {action, computed, observable} from "mobx";
import {JobTypeCode, PostalRegex} from "./_FieldConstants";
import * as moment from "moment";

export type FrameStatusCode = "S" | "F" | "L" | "U" | "D" | "E" | "N";
export type RawData = { [key: string]: string };

@model({
    transportLayer: "/api/orders",
    permissionTarget: "Order",
    routingInfo: {
        list: '/orders',
        create: '/orders/create',
        read: '/orders/:id/read',
    }
})
export class OrderModel extends Model {
    static types: Types = {
        raw_data: "json",
        date_ordered: "date",
    }

    constructor() {
        super();
        super.setup(); // This registers the observable attributes and viewModel
    }

    @observable readOnly = false;

    @attribute({label: t`Order Number`, description: t`The unique identifier for the order. This number is automatically generated`})
    order_number = "";

    @attribute({label: t`Job Type`, description: t`What this job is producing`})
    job_type: JobTypeCode = "P";

    @attribute({label: t`Customer Num`, description: t`The account this order is being made under`})
    customer_num = "";

    @attribute({label: t`Patient Name`, description: t`The name of the patient this order is for`, validation: string().required(t`Patient name is required`)})
    customer_name = "";

    @attribute({label: t`Customer City`, description: t`The shipping city for this patient`})
    customer_city = "";

    @attribute({label: t`Cust Job Id`})
    cust_job_id = "";

    @attribute({label: t`Date Completed`, description: t`The date this order was completed`})
    date_completed = "";

    @attribute({label: t`Date Received`, description: t`The date this order was received`})
    date_received = "";

    @attribute({label: t`Date Ordered`, description: t`The date this order was made`})
    date_ordered = moment();

    @attribute({label: t`Tray`, description: t`The tray number this order is associated with`})
    tray = "";

    @attribute({label: t`Consult Number`, description: t`Number used for external tracking`, validation: string().when('readOnly', {
            is: true,
            then: schema => schema.optional(),
            otherwise: schema => schema.required(t`Consult Number is a required field`)
        })})
    po_number = "";

    @attribute({label: t`Ship To Num`})
    ship_to_num = "";

    @attribute({label: t`Ship To Name`, description: t`Name to print on the shipping label`})
    ship_to_name = "";

    @attribute({label: t`Ship To Address1`, description: t`Address to print on the shipping label. Cannot contain special characters except period(.) or dash(-)`})
    ship_to_address1 = "";

    @attribute({label: t`Ship To Address2`, description: t`Address to print on the shipping label. Cannot contain special characters except period(.) or dash(-)`})
    ship_to_address2 = "";

    @attribute({label: t`Ship To Address3`, description: t`Address to print on the shipping label. Cannot contain special characters except period(.) or dash(-)`})
    ship_to_address3 = "";

    @attribute({label: t`Ship To Address4`, description: t`Address to print on the shipping label. Cannot contain special characters except period(.) or dash(-)`})
    ship_to_address4 = "";

    // @ts-ignore
    @attribute({
        label: t`Ship To Code`,
        description: t`Zip or postal code`,
        validation: string().when('readOnly', {
            is: true,
            then: schema => schema.optional(),
            otherwise: schema => schema.matches(PostalRegex,  { message: t`Zip Code/Postal Code must be of valid format e.g., 12345, A1A 1A1, B2B2B2`, excludeEmptyString: true })
        })})
    ship_to_code = "";

    @attribute({label: t`Shipping Method`, description: t`The method used for shipping the order`, validation: string().when('readOnly', {
        is: true,
            then: schema => schema.optional(),
            otherwise: schema => schema.required(t`Shipping method is a required field`)
        })})
    shipping_methods_id : string | null = "";

    @attribute({label: t`Last 4`, description: t`The last 4 digits of the Veteran's Social Security number`, validation: string().when('readOnly', {
            is: true,
            then: schema => schema.optional(),
            otherwise: schema => schema.length(4, t`Last 4 must be exactly 4 characters`)
        })})
    release_number = "";

    @attribute()
    raw_data: RawData = {};

    @attribute({label: t`Type`, description: t`The type of the order`, validation: string().when('job_type', {
        is: (jobType : string) => ["F", "A"].includes(jobType),
        then: schema => schema.nullable().optional(),
        otherwise: schema => schema.required(t`Type is a required field`)
    })})
    style_type_id: string | null = "";

    @attribute({label: t`Style`, description: t`The style of the order - dependent on the order type`, validation: string().when('job_type', {
        is: (jobType : string) => ["F", "A"].includes(jobType),
        then: schema => schema.nullable().optional(),
        otherwise: schema => schema.required(t`Style is a required field`)
    })})
    style_id: string | null = "";

    @attribute({label: t`Material`, description: t`The material used for the order - depending on the style`, validation: string().when('job_type', {
        is: (jobType : string) => ["F", "A"].includes(jobType),
        then: schema => schema.nullable().optional(),
        otherwise: schema => schema.required(t`Material is a required field`)
    })})
    style_material_id: string | null = "";

    @attribute({label: t`Frame Status`, description: t`What kind of frame are we sending with this order, if any`})
    _fstatus: FrameStatusCode | null = "U";

    @attribute({label: t`Addons`, description: t`What addons should be applies to this order`,
        validation: array()
            .ensure().max(15, t`Must have less than or equal to 15 addons`)
            .when('job_type', {
                is: (jobType : string) => jobType === "A" ,
                then: schema => schema.ensure().min(1, t`Must add at least one addon`),
                otherwise: schema => schema.ensure().min(0)
    })})
    addon_ids: string[] = [];

    @attribute({ label: t`Hold At Lab?`, description: t`Whether to hold the order at the lab or not` })
    hold_at_lab = false;

    @attribute()
    quick_code_id: string | null = "";

    @attribute({label: t`Customer`, validation: string().required()})
    customer_id = "";

    @attribute()
    trace_id: string | null = "";

    @attribute()
    style_type_name = "";

    @attribute()
    style_name = "";

    @attribute()
    style_material_name = "";

    @attribute()
    fupc: string | null = "";

    @attribute()
    fram: string | null = "";

    @attribute()
    _fmupc: string | null = "";

    @attribute({label: t`Submitted By`})
    submitted_by: string | null = "";

    @attribute({label: t`Frame SKU`, validation: string().nullable().when('_fstatus', {
            is: (frameStatus : string) => frameStatus === "S",
            then: schema => schema.required(t`Must select a frame from Rx-Inventory`),
            otherwise: schema => schema.optional()
        })})
    frame_sku: string | null = "";

    @attribute()
    unvalid_message: string | null = "";

    @attribute({validation: string().max(385).optional().nullable()})
    comments: string | null = "";

    @computed get isAddonJob() {
        return this.viewModel.job_type === "A";
    }

    @computed get isFrameOnlyJob() {
        return this.viewModel.job_type === "F";
    }

    @computed get isEdgeOnlyJob() {
        return this.viewModel.job_type === "E";
    }

    @computed get isInvalidJob() {
        if (this.unvalid_message){
         return this.unvalid_message.trim().length > 0;
        }
        return false
    }

    @computed get hasComment() {
        if (this.comments) {
            return this.comments.trim().length > 0;
        }
        return false;
    }

    @action.bound
    setReadOnly(readOnly: boolean) {
        this.readOnly = readOnly;
    }
}

export const OrderStore = OrderModel.store as unknown as DataStore<OrderModel>;

const tmpModel = OrderStore.createModel();
const tmpSchema = tmpModel.validationSchema

export const createOrderValidationSchema = () => {
    return object().shape({}).concat(tmpSchema);
}

// @ts-ignore
schemas["OrderModel"] = createOrderValidationSchema();